#ifdef __x86_64__
#error "this must be compiled for 32 bit"
#endif
#include "mboot.h"
#define MAX_LOG LOG_ERR
#include "../include/libce.h"
#include "../include/elf64.h"
#include "../include/vm.h"
#include "../include/boot.h"

// Il boostrap avviene in due stadi:
//
// STADIO 1 (realizzato dal bootloader incluso in QEMU)
// Carica in memoria il secondo bootloader e eventuali moduli. Il secondo boot
// loader deve essere un file ELF e viene caricato agli indirizzi specificati
// nella sua tabella di programma (normalmente nel secondo MiB), mentre i
// moduli possono essere file generici e vengono semplicemente copiati in RAM
// (subito dopo il secondo boot loader).  Il primo boot loader porta il
// processore da modo reale (16 bit) a modo protetto a 32 bit, quindi cede il
// controllo al secondo boot loader. Passa nel registro EAX una costante che
// permette al secondo boot loader di capire di essere stato avviato da un boot
// loader che rispetta lo standard "multiboot".  Nel registro EBX passa
// l'indirizzo di una struttura multiboot_info_t che contiene varie
// informazioni (memoria RAM installata nel sistema, indirizzi in RAM delle
// copie dei moduli, ...). La struttura è definita in "mboot.h" in questa
// directory.
//
// Stato della memoria alla fine dello stadio 1:
// - primo MiB: contiene la struttura multiboot_info_t
// - secondo MiB: contiene il secondo boot loader (caricato ai suoi indirizzi)
//    e i moduli semplicemente copiati in memoria
//
// STADIO 2 (realizzato da questo programma)
// Lo stadio 1 deve aver caricato almeno un modulo e questo modulo deve essere
// di tipo ELF. Il secondo bootloader interpreta questo modulo e copia i
// segmenti ELF al loro indirizzo di caricamento (che deve partire dal terzo
// MiB in poi per non sovrascrivere il boot loader stesso e i moduli). Porta
// poi la CPU nella modalità a 64 bit e cede il controllo al primo modulo.
// Passa in EDI il puntatore ad una struttura boot64_info che contiene varie
// informazioni (in particolare gli indirizzi dei moduli, che si trovano ancora
// in memoria dove erano stati caricati dal primo boot loader).
//
// Stato della memoria alla fine dello stadio 2
// - primo MiB: usato come heap dal secondo boot loader, contiene alcune
//   strutture dati che devono essere usate anche in seguito (GDT, TSS,
//   boot64_info, tabelle che compongono il TRIE della finestra FM) e
//   parti libere. L'indirizzo della prima parte libera viene passato
//   al primo modulo tramite boot64_info->memlibera in modo che lo heap
//   possa essere riutilizzato
// - secondo MiB: si trova nello stesso stato lasciato dal primo boot loader
// - terzo MiB: primo modulo caricato ai suoi indirizzi

// funzioni per l'allocazione e la deallocazione nello heap (primo MiB)
void *operator new(size_t s)
{
	return alloc(s);
}

void operator delete(void *p)
{
	dealloc(p);
}

void* operator new(size_t s, std::align_val_t a)
{
	return alloc_aligned(s, a);
}

// strutture dati che è necessario definire e inizializzare per il modo
// protetto a 64 bit. Qui definiamo delle strutture di cui ci interessa
// solo la dimensione e l'allineamento, per poterle allocare tramite new. La loro inizializzazione
// verrà fatta in assembler (si veda boot64/boot.s, funzioni gdt_init e tss_init).
struct GDT_entry {
	char bytes[DIM_GDT_ENTRY];
};

struct alignas(16) GDT {
	GDT_entry entry[NUM_GDT_ENTRIES];
};

struct alignas(16) TSS {
	char bytes[DIM_TSS];
};

extern "C" natq tss_init(TSS* tss);
extern "C" void gdt_init(GDT* gdt, TSS* tss);

// attiva la paginazione, definita in boot64/boot.s
extern "C" void attiva_paginazione(boot64_info *info, natl entry, int max_liv);

bool controlla_collisione(natq beg, natq end, multiboot_module_t* mod, natq num_moduli)
{
	for (natq i = 0; i < num_moduli; i++) {
		if (end <= mod[i].mod_start || beg > mod[i].mod_end)
			continue;
		flog(LOG_ERR, "Trovata sovrapposizione con il mod[%llu]", i);
		return false;
	}
	return true;
}

// carica il primo modulo secondo le indicazioni della sua tabella di programma
natl carica_modulo(multiboot_module_t* mod, natq num_moduli) {
	Elf64_Ehdr* elf_h = ptr_cast<Elf64_Ehdr>(mod->mod_start);

	// verifichiamo che si tratti di un file ELF
	if (!(elf_h->e_ident[EI_MAG0] == ELFMAG0 &&
	      elf_h->e_ident[EI_MAG1] == ELFMAG1 &&
	      elf_h->e_ident[EI_MAG2] == ELFMAG2 &&
	      elf_h->e_ident[EI_MAG2] == ELFMAG2))
	{
		flog(LOG_ERR, "Formato del modulo '%s' non riconosciuto",
			ptr_cast<char>(mod->cmdline));
		return 0;
	}

	// e che contenga un eseguibile per x86_64/amd64
	if (!(elf_h->e_ident[EI_CLASS] == ELFCLASS64  &&
	      elf_h->e_ident[EI_DATA]  == ELFDATA2LSB &&
	      elf_h->e_type	       == ET_EXEC     &&
	      elf_h->e_machine 	       == EM_AMD64))
	{
		flog(LOG_ERR, "Il modulo '%s' non contiene "
			"un esegubile per x86_64", ptr_cast<char>(mod->cmdline));
		return 0;
	}

	// interpretiamo la sua tabella di programma
	Elf64_Phdr* elf_ph = ptr_cast<Elf64_Phdr>(mod->mod_start + elf_h->e_phoff);
	for (int i = 0; i < elf_h->e_phnum; i++) {
		// ci interessano solo i segmenti PT_LOAD
		if (elf_ph->p_type != PT_LOAD)
			continue;

		// evitiamo di sovrascrivere quanto si trova già in memoria
		if (elf_ph->p_vaddr < 2*MiB) {
			flog(LOG_ERR, "Il modulo '%s' richiede un caricamento all'indirizzo '%llx',"
					"con possibile sovrascrittura di altri moduli",
					ptr_cast<char>(mod->cmdline), elf_ph->p_vaddr);
		}

		if (!controlla_collisione(elf_ph->p_vaddr,
				     elf_ph->p_vaddr + elf_ph->p_memsz,
				     mod, num_moduli))
			return 0;

		memcpy(voidptr_cast(elf_ph->p_vaddr),
		       voidptr_cast(mod->mod_start + elf_ph->p_offset),
		       elf_ph->p_filesz);
		flog(LOG_INFO, "- copiati %lld byte da %llx a %llx",
			elf_ph->p_filesz, mod->mod_start +  elf_ph->p_offset, elf_ph->p_vaddr);
		if (elf_ph->p_memsz != elf_ph->p_filesz) {
			memset((void*)(elf_ph->p_vaddr + elf_ph->p_filesz), 0,
			       elf_ph->p_memsz - elf_ph->p_filesz);
			flog(LOG_INFO, "- azzerati ulteriori %llu byte",
					elf_ph->p_memsz - elf_ph->p_filesz);
		}
		elf_ph = ptr_cast<Elf64_Phdr>(int_cast<uintptr_t>(elf_ph) +
			elf_h->e_phentsize);
	}
	flog(LOG_INFO, "- entry point %llx", elf_h->e_entry);
	return static_cast<natl>(elf_h->e_entry);
}

static const int MAXARGS = 10;
static int debug_mode = 0;

// scomponiamo la stringa cmd in argomenti separati da spazi.
// Al momento comprendiamo un solo argomento: '-s', che attiva
// la modalità debug.
static void parse_args(char *cmd)
{
	char *argv[MAXARGS];
	int argc = 0;

	char *scan = cmd;
	do {
		argv[argc] = scan;
		while (*scan && *scan != ' ')
			scan++;
		if (*scan) {
			*scan = '\0';
			for (scan++; *scan && *scan == ' '; scan++)
				;
		}
		argc++;
	} while (*scan && argc <= MAXARGS);

	for (int i = 0; i < argc; i++) {
		if (argv[i][0] != '-')
			continue;
		if (argv[i][1] == 's' && !debug_mode) {
			flog(LOG_INFO, "DEBUG MODE");
			debug_mode = 1;
		}
	}
}

extern "C" void panic(const char *)
{
}

// mappa la memoria fisica in memoria virtuale, inclusa l'area PCI
bool crea_finestra_FM(paddr root_tab, paddr mem_tot)
{
	auto identity_map = [] (vaddr v) -> paddr { return v; };
	// mappiamo tutta la memoria fisica:
	// - a livello sistema (bit U/S non settato)
	// - scrivibile (bit R/W settato)
	// - con pagine di grandi dimensioni (bit PS)
	//   (usiamo pagine di livello 2 che sono sicuramente disponibili)

	// vogliamo saltare la prima pagina per intercettare *nullptr, e inoltre
	// vogliamo settare il bit PWT per le pagine che contengono la memoria
	// video.  Per farlo dobbiamo rinunciare a settare PS per la prima regione
	natq first_reg = dim_region(1);

	// [0, DIM_PAGINA): non mappato
	// [DIM_PAGINA, 0xa0000): memoria normale
	if (map(root_tab, DIM_PAGINA, 0xa0000, BIT_RW, identity_map) != 0xa0000)
		return false;
	// [0xa0000, 0xc0000): memoria video
	if (map(root_tab, 0xa0000, 0xc0000, BIT_RW|BIT_PWT, identity_map) != 0xc0000)
		return false;
	// [0xc0000, first_reg): memoria normale
	if (map(root_tab, 0xc0000, first_reg, BIT_RW, identity_map) != first_reg)
		return false;

	// mappiamo il resto della memoria con PS settato
	if (mem_tot > first_reg) {
		if (map(root_tab, first_reg, mem_tot, BIT_RW, identity_map, 2) != mem_tot)
			return false;
	}

	flog(LOG_INFO, "Creata finestra sulla memoria centrale:  [%16llx, %16llx)", DIM_PAGINA, mem_tot);

	// Mappiamo tutti gli altri indirizzi, fino a 4GiB, settando sia PWT che PCD.
	// Questa zona di indirizzi è utilizzata in particolare dall'APIC per mappare i propri registri.
	vaddr	beg_pci = allinea(mem_tot, 2*MiB),
		end_pci = 4*GiB;
	if (map(root_tab, beg_pci, end_pci, BIT_RW|BIT_PCD|BIT_PWT, identity_map, 2) != end_pci)
		return false;

	flog(LOG_INFO, "Creata finestra per memory-mapped-IO:    [%16llx, %16llx)", beg_pci, end_pci);
	return true;
}

volatile int wait_for_gdb;
extern void* memlibera;
extern "C" void main (natl magic, multiboot_info_t* mbi)
{
	natl entry;
	boot64_info tmp_info;
	char *cmdline = nullptr;
	paddr mem_lower = 640*KiB;
	paddr mem_tot = 8*MiB;

	// inizializzazione della porta seriale per il debugging
	serial::init1();
	flog(LOG_INFO, "Boot loader di Calcolatori Elettronici, v1.0");

	// controlliamo di essere stati caricati da un bootloader che rispetti
	// lo standard multiboot
	if (magic != MULTIBOOT_BOOTLOADER_MAGIC) {
		flog(LOG_ERR, "Numero magico non valido: %#x", magic);
		return;
	}

	// se il primo bootloader ci ha passato informazioni sulla memoria, la usiamo
	// per costruire correttamente la finestra FM (altrimenti usiamo valori di default)
	if (mbi->flags & MULTIBOOT_INFO_MEMORY) {
		mem_lower = (mbi->mem_lower + 1) * KiB;
		mem_tot = (mbi->mem_upper + 0x480) * KiB;
	}
	flog(LOG_INFO, "Memoria totale: %lld MiB, heap: %lld KiB", mem_tot/MiB, (mem_lower - DIM_PAGINA)/KiB);

	// se il primo bootloader ci ha passato una stringa di opzioni, proviamo a esaminarla
	if (mbi->flags & MULTIBOOT_INFO_CMDLINE) {
		cmdline = ptr_cast<char>(mbi->cmdline);
		flog(LOG_INFO, "Argomenti: %s", cmdline);
		parse_args(cmdline);
	}

	// Il numero dei moduli viene passato dal primo bootloader in mods_count
	// Abbiamo bisogno di almeno un modulo.
	if (!(mbi->flags & MULTIBOOT_INFO_MODS) || mbi->mods_count < 1)
	{
		flog(LOG_ERR, "Informazioni sui moduli assenti o errate");
		return;
	}

	// raccogliamo le informazioni che ci servono su tutti i moduli
	memset(&tmp_info, 0, sizeof(tmp_info));
	flog(LOG_INFO, "Il boot loader precedente ha caricato %d modul%s:", mbi->mods_count,
			mbi->mods_count > 1 ? "i" : "o");
	multiboot_module_t* mod = ptr_cast<multiboot_module_t>(mbi->mods_addr);
	for (unsigned int i = 0; i < mbi->mods_count; i++) {
		flog(LOG_INFO, "- mod[%d]: start=%x end=%x file=%s",
			i, mod[i].mod_start, mod[i].mod_end, ptr_cast<const char>(mod[i].cmdline));
		// ignoriamo quelli oltre MAX_MODULES
		if (i < MAX_MODULES) {
			tmp_info.mod[i].mod_start = mod[i].mod_start;
			tmp_info.mod[i].mod_end = mod[i].mod_end;
		}
	}
	// e carichiamo il primo
	flog(LOG_INFO, "Copio mod[0] agli indirizzi specificati nel file ELF:");
	entry = carica_modulo(mod, mbi->mods_count);
	if (!entry) {
		flog(LOG_ERR, "ATTENZIONE: impossibile caricare mod[0]");
		return;
	}

	// ora mbi non ci serve più e possiamo riusare la memoria nel primo MiB come heap
	heap_init(voidptr_cast(DIM_PAGINA), mem_lower - DIM_PAGINA);
	// nello heap creiamo tre strutture che devono essere preservate anche dopo che
	// avremo ceduto il controllo al primo modulo:
	// 1) La struttura boot64_info;
	boot64_info *info = new boot64_info{tmp_info};
	if (!info) {
		flog(LOG_ERR, "ATTENZIONE: impossibile allocare boot64_info");
		return;
	}
	// 2) Il segmento TSS
	TSS *tss = new TSS;
	if (!tss) {
		flog(LOG_ERR, "ATTENZIONE: impossibile allocare il segmento TSS");
		return;
	}
	// che inizializziamo, passando al primo modulo l'informazione su dove
	// si trova tss_punt_nucleo al suo interno;
	info->tss_punt_nucleo = tss_init(tss);
	// 3) la GDT
	GDT *gdt = new GDT;
	if (!gdt) {
		flog(LOG_ERR, "ATTENZIONE: impossible allocare la tabella GDT");
		return;
	}
	// che inizializziamo.
	gdt_init(gdt, tss);
	// i dettagli su GDT e TSS si trovano in boot64/boot.s

	// Ora creiamo la finestra FM. Le tabelle verranno allocate nello heap.
	paddr root_tab = alloca_tab();
	if (!root_tab) {
		flog(LOG_ERR, "ATTENZIONE: impossibile allocare la tabella radice");
		return;
	}
	if (!crea_finestra_FM(root_tab, mem_tot)) {
		flog(LOG_ERR, "ATTENZIONE: fallimento in crea_finestra_FM()");
		return;
	}

	// prima di attivare la paginazione dobbiamo caricare cr3
	loadCR3(root_tab);
	if (debug_mode) {
		wait_for_gdb = 1;
		flog(LOG_INFO, "Attendo collegamento da gdb...");
	}
	// Da questo punto in poi non ci sono più allocazioni nello heap. Passiamo
	// l'attuale contenuto di memlibera (testa della lista dei chunk liberi
	// nello heap) al primo modulo, in modo che possa usare il resto della
	// memoria disponibile nel primo MiB.
	info->memlibera = int_cast<uintptr_t>(memlibera);
	flog(LOG_INFO, "Attivo la modalita' a 64 bit e cedo il controllo a mod[0]...");
	attiva_paginazione(info, entry, MAX_LIV);

	/* mai raggiunto */
	return;
}
