// boot.s
#include "../include/libce.h"
#include "mboot.h"

/////////////////////////////////////////////////////////////////////////
// AVVIO                                                                //
//////////////////////////////////////////////////////////////////////////
// Il primo bootstrap loader (in QEMU) attiva il modo protetto (per poter
// accedere agli indirizzi di memoria principale superiori a 1MiB) e carica il
// secondo boostrap loader (questo programma) in memoria, quindi salta alla
// sua prima istruzione. Il primo bootstrap loader può anche passare delle
// informazioni al secondo (tramite i registri e la memoria).
//
// In questo sistema usiamo lo standard multiboot, che definisce il formato che
// il secondo loader deve rispettare e il formato delle informazioni passate
// dal primo loader al secondo.  Il formato del file contenente il secondo
// loader deve essere quello di un normale file eseguibile, ma, nei primi
// 2*4KiB, deve contenere la struttura multiboot_header, definita più avanti.
// Il primo loader, prima di saltare alla prima istruzione del secondo (l'entry
// point specificato nel file eseguibile), lascia nel registro %eax un valore
// di riconoscimento e in %ebx l'indirizzo di una struttura dati, contentente
// varie informazioni (in particolare, la quantità di memoria principale
// installata nel sistema, il dispositivo da cui è stato eseguito il bootstrap
// e l'indirizzo di memoria in cui sono stati caricati gli eventuali moduli)
.text

#define STACK_SIZE                      0x4000
#define MULTIBOOT_HEADER_FLAGS         0x00000003

.globl  _start, start
_start:				// entry point
start:
     jmp     multiboot_entry	// scavalchiamo la struttra richiesta
				// dal primo loader, che deve
				// trovarsi verso l'inizio del file
.align  4
// le seguenti informazioni sono richieste dal primo loader
multiboot_header:
     .long   MULTIBOOT_HEADER_MAGIC 				// valore magico
     .long   MULTIBOOT_HEADER_FLAGS   				// flag
     .long   -(MULTIBOOT_HEADER_MAGIC + MULTIBOOT_HEADER_FLAGS) // checksum
     .long   multiboot_header
     .long   _start
     .long   _edata
     .long   _end
     .long   multiboot_entry

multiboot_entry:

     cli
     // cambiamo lo stack perché riutilizzeremo come parte dello heap
     // la memoria in cui si trova lo stack attuale
     movl    $stack_end, %esp

     pushl   %ebx			// parametri passati dal primo loader
     pushl   %eax
     call    disable_8259
     call    main			// il resto dell'inizializzazione
					// è scritto in C++
     call    reboot			// se main ritorna c'è stato un errore

//////////////////////////////////////////////////////////////////
// funzioni di utilità						//
//////////////////////////////////////////////////////////////////
//
// Per abilitare la modalità a 64 bit con i due livelli utente/sistema dobbiamo
// definire e inizializzare due strutture dati: la GDT e almeno un segmento
// TSS. Entrambe risalgono al 286 e al suo sofisticato sistema di
// segmentazione:
// - La GDT (Global Descriptor Table) è una tabella di descrittori di segmento;
// ogni segmento è definito da una base, un limite e alcuni attributi. I
// segmenti identificano un intervallo di indirizzi [base, base+limite] e i
// loro attributi ne descrivono il contenuto (codice, dati, ...), le
// limitazioni sull'utilizzo (il livello di privilegio, la possibilità di
// scrivervi o meno, etc.) o danno altre informazioni (in particolare la
// dimensione degli operandi nel caso di segmenti di tipo codice).
// - I segmenti TSS (Task State Segment) sono un particolare tipo di segmento
// che nelle intenzioni dell'Intel doveva essere usato al posto dei descrittori
// di processo per realizzare i cambi di contesto interamente in hardware.
//
// I segmenti TSS ci interessano solo perché contengono il puntatore alla pila
// sistema utilizzato dal meccanismo delle interruzioni, e per disabilitare
// un certo meccanismo che interferisce con l'IOPL.
//
//  La GDT ci interessa per tre motivi:
//  - ci permette di avere codice con operandi a 64 bit (vedere il flag L più
//    avanti);
//  - ci permette di avere codice con livelli di privilegio diversi
//    per l'utente e il sistema (vedere campo DPL più avanti);
//  - deve contenere un descrittore del segmento TSS di cui sopra.
//
// Come i gate della IDT, anche i descrittori della GDT sono molto più
// complicati di quanto avrebbero potuto essere, per via della loro storia:
// - nel 286 la base era su 24 bit e il limite su 16 bit
//   (segmenti di max 64 KiB ciascuno in uno spazio di memoria da 16 MiB).
//   I descrittori occupavano 6 byte.
// - nel 386 la base è passata a 32 bit e il limite a 20 bit, con un
//   flag di granularità (flag G) che permetteva di interpretare il
//   limite come numero di byte (G=0) o come numero di pagine da 4 KiB (G=1)
//   (segmenti di max 1 MiB o 4 GiB in uno spazio di memoria da 4 GiB).
//   I descrittori occupavano 8 byte.
// - l'architettura a 64 bit dell'AMD ha reso base e limite non più
//   significativi per la maggior parte dei segmenti. Il descrittore
//   di segmento TSS è uno di quelli che fa eccezione, e in quel caso
//   il descrittore è stato esteso a 16 byte per poter avere una base
//   a 64 bit.
//
// Il formato odierno dei descrittori è il seguente (la seconda riga
// può essere assente se il tipo non la richiede):
//
//
//       ____________ base[24-31]     `
//      /                             |
//      |       _____ flags           | aggiunti nel 386
//      |      /                      |
//      |     |    __ limit[16-19]    /
//      |     |   /
//      |     |   |
//      |     |   |          descrittore originale del 286
//      |     |   | |-----------------------------------------------|
//   63 |     |   |  47   40 39   32 31           16 15            0
//  +---+---+-+-+-+-+---+---+---+---+---+---+---+---+---+---+---+---+
//  |   .   | . | . | access|     base[0-23]        |  limit[0-15]  |  +0
//  +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+       `
//  | -   -   -   -   -   -   -   - |      base[32-63]              |  +8   | amd64
//  +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+       /
//        +7      +6      +5      +4      +3      +2      +1
//
//  L'access byte ha il seguente formato (sostanzialmente mai cambiato):
//
//    7   6   5   4   3   2   1   0           P: presenza
//  +---+---+---+---+---+---+---+---+         DPL: Desc. Priv. Level
//  | P |  DPL  | S |     Type      |         S: codice/dati (1) vs sistema (0)
//  +---+---+---+---+---+---+---+---+         Type: tipo
//
//  E questi sono i flag:
//
//    3   2   1   0                           G: granularità	(386)
//  +---+---+---+---+                         DB: 16/32 bit	(386)
//  | G |DB | L | - |                         L: 64 bit 	(amd64)
//  +---+---+---+---+
//
// Nell'archiettura 286/386, ad ogni istante sono attivi solo un numero
// limitato di segmenti: quelli selezionati dai registri selettori di segmento
// (CS, DS, SS, etc.).  Questi registri vanno caricati con dei "selettori di
// segmento", che sono quantità di 16 bit con il seguente formato:
//
// Segment selector:
//
//   15                                              3       1   0
// +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
// |                indice nella GDT                   |   |  RPL  |
// +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
//
// (RPL: Requested Privilege Level)
//
// Quando si carica un selettore in uno dei registri CS, DS, SS etc.  il
// processore 286/386 esegue vari controlli sul tipo e gli attributi del segmento
// indicizzato, il privilegio corrente, l'RPL e il DPL del segmento
// indicizzato. Per esempio, CS è il Code Selector e può essere caricato
// solo con segmenti codice, DS è il Data Selector e può essere caricato
// solo con segmenti dati, SS è lo Stack Selector e può essere caricato
// solo con segmenti dati scrivibili.
// Il "privilegio corrente" non è altro che l'RPL del descrittore di
// segmento codice che si trova attualmente in CS. Per quel che ci
// riguarda, è sempre uguale al DPL del descrittore di segmento codice
// indicizzato dal selettore.

// I controlli sui privilegi e sugli attributi in generale sono stati quasi
// completamente eliminati nell'amd64, ma non del tutto. CS deve contenere il
// selettore di segmento codice e l'RPL di questo selettore (uguale, per quanto
// detto prima, al DPL del segmento) continua a identificare il livello di
// privilegio corrente. DS e SS non sono usati a livello sistema e possono
// essere caricati con 0, ma la IRETQ controlla che SS venga caricato con
// qualcosa di valido quando si sta tornando a livello utente. Per essere
// valido, l'RPL caricato in SS deve essere uguale al livello di privilegio
// corrente.  Per avere due livelli di privilegio ci servono dunque due
// segmenti codice, uno con DPL=0 (sistema) e uno con DPL=3 (utente).  Inoltre,
// ci serve un segmento dati con DPL=3 per poterlo caricare in SS quando il
// processore è a livello utente.
//
//  Inizializzeremo dunque la GDT nel seguente modo:
// (si vedano anche le #define in fondo a libce.h: gli ID_* sono gli indici
//  nella GDT e i SEL_* sono i corrispondenti selettori)
//
//  indice    offset   selettore  contenuto
//                     (base 16)
//  ==============================================
//        0        0	00	  descrittore nullo
//	  1	   8    08	  descrittore segmento codice sistema
//	  2	  16    13	  descrittore segmento codice utente
//	  3	  24    1b	  descrittore segmento pila/dati utente
//	  4	  32    20 	  descrittore segmento TSS
//		  40    	  (parte alta del descrittore segmento TSS)
// dimensione: 48 byte
//
// Nota: il descrittore di indice 0 deve essere nullo.

.global gdt_init
gdt_init:
	push %ebp
	mov %esp, %ebp
	push %edi

	mov 8(%ebp), %edx		// &gdt -> edx
	// aggiorniamo il puntatore alla gdt che verra usato da
	// attiva_paginazione (istruzione lgdt)
	mov %edx, gdt64_ptr + 2
	// azzeriamo tutta la gdt
	  mov %edx, %edi
	  xor %eax, %eax
	  mov $((DIM_GDT_ENTRY*NUM_GDT_ENTRIES)>>2), %ecx
	  rep stosl
	// descrittore segmento codice sistema
	lea (ID_CODE_SYS*8)(%edx), %edi
	movb $0b10011011, 5(%edi)	// P=1, DPL=sys, S=cod/dati, tipo=codice
	movb $0b00100000, 6(%edi)	// L=1 (64 bit)
	// descrittore segmento codice utente
	lea (ID_CODE_USR*8)(%edx), %edi
	movb $0b11111011, 5(%edi)	// P=1, DPL=usr, S=cod/dati, tipo=codice
	movb $0b00100000, 6(%edi)	// L=1 (64 bit)
	// descrittore segmento pila/dati utente
	lea (ID_DATA_USR*8)(%edx), %edi
	movb $0b11110011, 5(%edi)	// P=1, DPL=usr, S=cod/dati, tipo=dati scrivibile
					// (l'amd64 ignora l'attributo scrivible, ma lo
					// settiamo lo stesso per documentazione)
	// descrittore segmento TSS
	// qui base e limite sono significativi e devono corrispondere
	// all'indirizzo e alla dimensione del segmento TSS
	lea (ID_SYS_TSS*8)(%edx), %edi
	movl 12(%ebp), %eax		// &tss -> eax
	movw %ax, 2(%edi)		// base[0-15]
	shrl $16, %eax
	movb %al, 4(%edi)		// base[16-23]
	movb %ah, 7(%edi)		// base[24-31]
	// non scriviamo in base[61-32] perché il nostro TSS è sicuramente
	// nella parte bassa della memoria
	movw $(DIM_TSS-1), (%edi)	// limit (sizeof(TSS)-1)
	movb $0b10001001, 5(%edi)	// P=1, DPL=sys, S=sistema, tipo=TSS

	pop %edi
	leave
	ret

// Anche la strana struttura del TSS è dovuta alla sua storia. Il TSS del
// 386 inizia così
//
//      +3  +2  +1  +0
//    +---+---+---+---+
//    |   |   |   |   |  0
//    +---+---+---+---+
//    | - | - |  SS0  | +4
//    +---+---+---+---+
//    |     ESP0      | +8
//    +---+---+---+---+
//           ...
//
// Dove SS0:ESP0 è il puntatore alla pila di livello 0 (sistema) espressa come
// selettore di segmento (16 bit) e offset (32 bit), in accordo con
// l'architettura segmentata.  Nell'amd64 serve un puntatore alla pila non
// segmentato e su 64 bit. L'AMD ha dunque riutilizzato i 4 byte all'offset +4
// per estendere ESP0 in RSP0 (64 bit): il risultato è che RSP0 nel segmento TSS
// della modalità a 64 bit è disallineato:
//
//     +7  +6  +5  +4  +3  +2  +1  +0
//   +---+---+---+---+---+---+---+---+
//   | RSP0 (low)    |   |   |   |   |  0
//   +---+---+---+---+---+---+---+---+
//        ...        |   RSP0 (high) | +8
//                   +---+---+---+---+
//
// Inoltre, essendo sparito SS0, il meccanismo delle interruzioni carica un valore
// costante in SS (il selettore nullo).
//
// Gli ultimi due byte del TSS contengno l'offset (a partire dalla base del
// TSS) di una "iobitmap". Questa struttura dati contiene un bit per ogni
// indirizzo dello spazio di I/O (65536 bit, pari a 2 KiB).  Se l'iobitmap è
// presente, il processore concede l'accesso alle porte di I/O se il
// corrispondente bit è zero, anche se il privilegio corrente è (logicamente)
// inferiore dell'IOPL.  Nell'idea originaria il sistema dovrebbe definire un
// segmento TSS per ogni processo e in questo modo la iobitmap avrebbe permesso
// di definire con precisione quali processi possono accedere a quali porte di
// I/O.  Nel nostro caso vogliamo definire un unico TSS e usare l'IOPL per
// vietare completamente l'accesso allo spazio di I/O ai processi utente, ma
// per farlo dobbiamo "disattivare" il meccanismo dell'I/O bitmap. Questo si
// ottiene definendo un offset che punti oltre il limite del TSS stesso. Dal
// momento che il TSS è grande 104 byte, useremo un offset di 104.

.global tss_init
tss_init:
	push %ebp
	mov %esp, %ebp
	push %edi
	mov 8(%ebp), %edx	// &tss -> edx
	// azzeriamo tutto il TSS
	  mov %edx, %edi
	  xor %eax, %eax
	  mov $(DIM_TSS>>2), %ecx
	  rep stosl
	// scriviamo la dimensione del TSS nell'offset della iobitmap
	movw $DIM_TSS, (DIM_TSS - 2)(%edx)
	// restituiamo l'indirizzo del campo RSP0
	lea 4(%edx), %eax
	xor %edx, %edx
	pop %edi
	leave
	ret

.global attiva_paginazione
// attiva la paginazione
attiva_paginazione:
	// dobbiamo attivare PAE (Physical Address Extensions) settando il bit
	// 4 di CR4. PAE è stato introdotto nel Pentium Pro per permettere
	// l'accesso a più di 4 GiB di memoria fisica. Con PAE la dimensione
	// dei descrittori di pagina e tabella passa da 4 a 8 byte.
	movl %cr4, %eax
        orl $0x00000020, %eax    // cr4.pae
	// opzionalmente abilitiamo anche LA57 settando il bit 12 di CR4.
	// Con LA57 i bit significativi degli indirizzi virtuali passano da 48
	// a 57 e i TRIE passano da 4 a 5 livelli (non tutti i processori
	// supportano LA57).
	cmpl $4, 12(%esp)
	je 1f
	orl $0x00001000, %eax	 // cr4.la57
1:	movl %eax, %cr4

	// Dal Pentium in poi processori Intel/AMD implementano una serie di
	// MSR (Model Specific Register) a cui si accede con le istruzioni
	// rmdsr (lettura) e wrmsr (scrittura). Ogni MSR è identificato da un
	// codice numerico che va passato tramite ECX. Qui ci interessa il
	// registro EFER (Extended Feature Enable Register, codice 0xC0000080)
	// introdotto dall'AMD nel K7.  Il bit 15 di questo registra abilita
	// LME (Long Mode Extension), ovvero la modalità a 64 bit. Attenzione
	// però: la modalità a 64 bit è una sotto-modalità della paginazione, e
	// quindi verrà effettivamente attivata solo quando avremo attivato
	// anche la paginazione
        movl $0xC0000080, %ecx   // efer msr
        rdmsr
        orl $0x00000100, %eax    // efer.lme
        wrmsr

	// Per attivare la paginazione settiamo il bit 31 di CR0.
	// Visto che abbiamo abilitato anche PAE e settato il bit 15 di EFER,
	// il processore passerà contestualmente nella modalità a 64 bit.
        movl %cr0, %eax
        orl $0x80000000, %eax    // cr0.pg
        movl %eax, %cr0
	// ora la modalità a 64 bit è attiva, ma il segmento corrente usa
	// ancora operandi a 32 bit. Dobbiamo caricare in CS il selettore del
	// segmento di codice sistema che aveamo preparato nella GDT. Prima
	// però dobbiamo caricare nel processore l'indirizzo della nostra GDT
	// (istruzione lgdt):
	lgdt gdt64_ptr
	// Non possiamo caricare CS con una movw, ma possiamo eseguire un "far
	// jump", cioè un salto in cui si specifica un selettore di segmento
	// oltre all'offset di destinazione
	ljmp $SEL_CODICE_SISTEMA, $long_mode

	// l'assemblatore sta assemblando questo file assumendo che si tratti
	// di codice a 32 bit. Con ".code64" gli chiediamo di assemblare il
	// resto con istruzioni a 64 bit
.code64
long_mode:
	// azzeriamo tutti i flag (tranne il bit 1 che per motivi storici deve
	// essere sempre settato)
	pushq $2
	popfq
	// azzeriamo tutti i gli altri registri selettori di segmento
	movw $0, %ax
	movw %ax, %ds
	movw %ax, %ss
	movw %ax, %fs
	movw %ax, %gs
	movw %ax, %es
	// carichiamo il selettore del segmento TSS nel registro speciale TR.
	// Il meccanismo delle interruzioni usa questo registro per accedere al
	// campo del TSS che contiene l'indirizzo della pila sistema.
	movw $SEL_SYS_TSS, %ax
	ltr %ax

	// se era stata richiesta la modalità di debug aspettiamo che gdb si
	// colleghi. Lo script 'debug' usa gdb per scrivere 0 in wait_for_gdb,
	// in modo da fare uscire il boot loader da questo ciclo e fargli
	// cedere il controllo al primo modulo.
wait:
	cmpl $1, wait_for_gdb(%rip)
	pause
	je wait

	// cediamo il controllo al primo modulo
	movl 4(%rsp), %edi	// boot64_info*
	movl 8(%rsp), %eax	// entry point
	movl $stack_end, %esp	// stack allineato a 16
	jmp *%rax

//////////////////
// sezione dati //
//////////////////
.code32
.data
// gdt64_ptr è usato dall'istruzione lgdt per caricare il registro GDTR con la
// base e il limite della GDT. Il limite si trova nei primi due byte e la base
// nei successivi 4 byte.
gdt64_ptr:
	.word (DIM_GDT_ENTRY*NUM_GDT_ENTRIES)-1
	.long 0		// verrà impostato a runtime
.align 16
	.global stack
stack:
	.space STACK_SIZE, 0
stack_end:
