/// @cond INTERNAL
#include "include/libce.h"
#include "include/vm.h"

static const uintmax_t UINT_MAX = 4294967295U;
// inizializza la idt per i programmi "bare" (esempiIO e
// esercizi di traduzione)
extern "C" void init_idt();
// restituisce il puntatore al primo 'c' dentro la stringa 'src',
// o nullptr se non trovato
char *index(const char *src, int c);
// come strlen, ma confronta al massimo 'len' caratteri
size_t strnlen(const char *s, size_t len);
// disabilita le interruzioni esterne mascherabili
extern "C" void cli();
// abilita le interruzioni esterne mascherabili
extern "C" void sti();
// scrive 'msg' sul log e spegne la macchina
extern "C" [[noreturn]] void panic(const char *msg);
// invoca i costruttori definiti nella sezione init_array
extern "C" void ctors();
// invoca i distruttori definiti nella sezione fini_array
extern "C" void dtors();
// formatta in base 'fmt' e scrivendo nel buffer 'str'
int vsnprintf(char *str, size_t size, const char *fmt, va_list ap);

// Manipolazione dei gate della IDT
// ================================
//
// I gate della IDT hanno una struttura complessa, dovuta alla loro storia:
// i gate sono stati introdotti nell'architettura a 16 bit del 286 con la
// seguente forma, su 6 byte (3 word):
//
//          +5  +4  +3  +2  +1  +0
// +---+---+---+---+---+---+---+---+
// | xxxxx | A | 0 |  sel  |  off  |
// +---+---+---+---+---+---+---+---+
//
// Dove A è il byte di accesso e sel:off è l'indirizzo della routine
// nello spazio segmentato (selettore di segmento + offset all'interno
// del segmento).
//
// Il byte di accesso A ha la seguente forma:
//
//   7 6 5 4 3 2 1 0
//  +-+-+-+-+-+-+-+-+
//  |P|DPL|0| tipo  |
//  +-+-+-+-+-+-+-+-+
//
// i tipi che ci interessano sono: 1110 (interrupt) oppure 1111
// (trap), ma il 286 ne definiva anche altri.
//
// Nel 386 l'offset è stato esteso a 32bit e il descrittore è stato
// portato a 8 byte. I 16 bit in più dell'offset sono stati aggiunti
// in cima e il resto è rimasto uguale:
//
//  +7  +6  +5  +4  +3  +2  +1  +0
// +---+---+---+---+---+---+---+---+
// | o31:16| A | 0 |  sel  | o15:0 |
// +---+---+---+---+---+---+---+---+
//
// Per l'architettura a 64 bit, AMD ha portato l'offset a 64 bit e il
// descrittore a 16 byte. L'architettura non è più segmentata, ma 'sel' rimane
// per specificare il livello di privilegio e la modalità a 64 bit, mentre
// 'offset' è l'indirizzo virtuale della routine. Il byte di accesso continua
// ad avere la stessa forma, ma i tipi ammessi sono solo interrupt (1110) o
// trap (1111).
//
//  +7  +6  +5  +4  +3  +2  +1  +0
// +---+---+---+---+---+---+---+---+
// | o31:16| A | 0 |  sel  | o15:0 |
// +---+---+---+---+---+---+---+---+
// |       0       |     o63:32    |  +8
// +---+---+---+---+---+---+---+---+
//
// Manipoliamo questi campi direttamente in assembly. In C++ definiamo una
// struttura 'gate' che ha il solo scopo di occupare lo spazio necessario:
struct alignas(16) gate {
	natq qw1;
	natq qw2;
};
// la IDT è un array di questi 'gate'
extern gate idt[];
// la funzione componi_gate (scritta in assembly), si preoccupa di creare un
// gate in base ai valori di 'routine' (indirizzo virtuale della routine che
// vogliamo associare al gate), 'trap' (flag per scegliere tra gate di tipo
// interrupt o trap) e 'liv' (DPL, che deve essere 3 per livello utente o 0 per
// livello sistema). La funzione imposta 'sel' sempre nello stesso modo:
// la routine girerà a livello sistema in modalità 64bit.
extern "C" gate componi_gate(void routine(), bool trap, int liv);

/// Associa una stringa descrittiva ad ogni tipo di eccezione.
extern const char* eccezioni[];

namespace kbd {

	const natl MAX_CODE = 42;
	const ioaddr iRBR = 0x60;
	const ioaddr iTBR = 0x60;
	const ioaddr iCMR = 0x64;
	const ioaddr iSTR = 0x64;

	extern bool shift;
	extern natb tab[MAX_CODE];
	extern char tabmin[MAX_CODE];
	extern char tabmai[MAX_CODE];

	void init();
	void reset();
	// un piedino del microcontrollore della tastiera comanda la circuiteria
	// di reset, quindi la funzione reboot() è definita qui.
	void reboot();
}

namespace vid {

	const natl COLS = 80;
	const natl ROWS = 25;
	const natl VIDEO_SIZE = COLS * ROWS;
	const natb CUR_HIGH = 0x0e;
	const natb CUR_LOW = 0x0f;

	const ioaddr iIND = 0x03D4;
	const ioaddr iDAT = 0x03D5;

	extern volatile natw* video;
	extern natb x, y;
	extern natw attr;

	void cursor();
	void scroll();
}

namespace svga {

	extern volatile void* framebuffer;
	extern volatile natw* vgareg;
	extern volatile natw* vbeext;

	//indici per vbeext
	static const natw VBE_DISPI_INDEX_ID = 0x0;
	static const natw VBE_DISPI_INDEX_XRES = 0x1;
	static const natw VBE_DISPI_INDEX_YRES = 0x2;
	static const natw VBE_DISPI_INDEX_BPP = 0x3;
	static const natw VBE_DISPI_INDEX_ENABLE = 0x4;
	static const natw VBE_DISPI_INDEX_BANK = 0x5;
	static const natw VBE_DISPI_INDEX_VIRT_WIDTH = 0x6;
	static const natw VBE_DISPI_INDEX_VIRT_HEIGHT = 0x7;
	static const natw VBE_DISPI_INDEX_X_OFFSET = 0x8;
	static const natw VBE_DISPI_INDEX_Y_OFFSET = 0x9;

	//valori possibili per VBE_DISPI_INDEX_BPP
	static const natw VBE_DISPI_BPP_4 = 0x04;
	static const natw VBE_DISPI_BPP_8 = 0x08;
	static const natw VBE_DISPI_BPP_15 = 0x0F;
	static const natw VBE_DISPI_BPP_16 = 0x10;
	static const natw VBE_DISPI_BPP_24 = 0x18;
	static const natw VBE_DISPI_BPP_32 = 0x20;

	//bit settabili su VBE_DISPI_INDEX_ENABLE
	static const natw VBE_DISPI_DISABLED = 0x00;
	static const natw VBE_DISPI_ENABLED = 0x01;
	static const natw VBE_DISPI_GETCAPS = 0x02;
	static const natw VBE_DISPI_8BIT_DAC = 0x20;
	static const natw VBE_DISPI_LFB_ENABLED = 0x40;
	static const natw VBE_DISPI_NOCLEARMEM = 0x80;

	bool init();
}

namespace hd {

	const ioaddr iBR  = 0x01F0;
	const ioaddr iERR = 0x01F1;
	const ioaddr iSCR = 0x01F2;
	const ioaddr iSNR = 0x01F3;
	const ioaddr iCNL = 0x01F4;
	const ioaddr iCNH = 0x01F5;
	const ioaddr iHND = 0x01F6;
	const ioaddr iCMD = 0x01F7;
	const ioaddr iSTS = 0x01F7;
	const ioaddr iDCR = 0x03F6;
	const ioaddr iASR = 0x03F6;

	void set_lba(natl lba);
	void wait_for_br();

}

namespace bm {

	extern ioaddr iBMCMD;
	extern ioaddr iBMSTR;
	extern ioaddr iBMDTPR;

}

namespace pci {

	const ioaddr iCAP = 0x0CF8;
	const ioaddr iCDP = 0x0CFC;

	natl make_CAP(natb bus, natb dev, natb fun, natb off);
}

namespace apic {

	// numero di piedini per le richieste di interruzione
	const natl IRQ_MAX = 24;

	// l'APIC ha alcuni registri mappati nello spazio di memoria e diversi
	// registri interni. I registri IRR, ISR, EOIR, IOREGSEL e IOWIN sono mappati
	// nello spazio di memoria e vi si accede con normali istruzioni con
	// operandi in memoria (per es. mov). In C++ è sufficiente definire un
	// puntatore che contenga l'indirizzo del registro e poi dereferenziarlo.

	// puntatore a End Of Interrupt Register
	static volatile natl* const pEOIR = ptr_cast<volatile natl>(0xFEE000B0);
	// puntatore a I/O REGister Selector (indirizzo non noto staticamente)
	extern volatile natl* pIOREGSEL;
	// puntatore a I/O WINdow (indirizzo non noto staticamente)
	extern volatile natl* pIOWIN;

	// funzioni di utilità per leggere IRR e ISR. Ciascun registro occupa in realtà
	// 8 sotto-registri da 4 byte, con indice i da 0 a 7.
	natl read_irr(natb i);
	natl read_isr(natb i);

	// I registri interni sono tutti di 4 byte. Vi si può accedere solo
	// indirettamente, tramite IOREGSEL e IOWIN: prima si scrive l'indice
	// del registro interno desiderato in IOREGSEL e a quel punto IOWIN
	// funge da finestra sul registro interno: leggendo e scrivendo IOWIN
	// si legge e scrive il registro interno selezionato.  Le funzioni in()
	// e out() semplificano questo procedimento.
	//
	// leggi il registro interno di indice 'off'
	natl in(natb off);
	// scrivi 'v' nel registro interno di indice 'off'
	void out(natb off, natl v);

	// i registri interni più importanti per il programmatore sono
	// quelli che compongono la "redirection table", che inizia all'indice 0x10.
	const natl RTO = 0x10;
	// La tabella ha 24 righe (una per ogni piedino di richiesta di
	// interruzione) e ogni riga occupa due registri interni (detti h e l)
	//
	//     h        l
	// +--------+--------+
	// |        |        |  RTO           (IRQ 0)
	// +--------+--------+
	// |        |        |  RTO + 2       (IRQ 1)
	// +--------+--------+
	// ~                 ~
	// +-----------------+
	// |        |        |  RTO + 2*23    (IRQ 23)
	// +--------+--------+
	//
	// La parte 'l' di ogni riga contiene le informazioni più importanti:
	// il vettore associato all'IRQ (bit 0-7), il bit per mascherare
	// le richieste (bit 16) e il bit che seleziona il Trigger Mode
	// sul livello oppure sul fronte (bit 15).

	const natl TRGM_BIT = (1U << 15); // trigger mode (1=livello, 0=fronte)
	const natl MIRQ_BIT = (1U << 16); // maschera (1=mascherato)

	// funzioni di utitilità per leggere/scrivere la parte h o l
	// della riga della redirection table relativa all'irq specificato
	natl read_rth(natb irq);
	natl read_rtl(natb irq);
	void write_rth(natb irq, natl w);
	void write_rtl(natb irq, natl w);

}

// Lo heap è composto da zone di memoria libere e occupate. Ogni zona di memoria
// è identificata dal suo indirizzo di partenza e dalla sua dimensione.
// Ogni zona di memoria contiene, nei primi byte, un descrittore di zona di
// memoria (di tipo des_mem) con un campo "dimensione" (dimensione in byte
// della zona, escluso il descrittore stesso) e un campo "next", significativo
// solo se la zona è libera, nel qual caso il campo punta alla prossima zona
// libera. Si viene quindi a creare una lista delle zone libere, la cui testa
// è puntata dalla variabile "memlibera" (allocata staticamente). La lista è
// mantenuta ordinata in base agli indirizzi di partenza delle zone libere.

// descrittore di zona di memoria: descrive una zona di memoria nello heap di
// sistema
struct alignas(sizeof(natq)) des_mem {
	natq dimensione;
	natq next;
};

// testa della lista delle zone libere
extern des_mem* memlibera;

static inline des_mem* natq2desmem(natq p)
{
	return ptr_cast<des_mem>(p);
}

static inline natq desmem2natq(des_mem *p)
{
	return int_cast<natq>(p);
}

bool free_interna(void* indirizzo, size_t quanti);

namespace timer {

	static const ioaddr iCWR = 0x43;

	static const ioaddr iCTR0_LOW = 0x40;
	static const ioaddr iCTR0_HIG = 0x40;
	static const ioaddr iSTS0_LOW = 0x40;
	static const ioaddr iSTS0_HIG = 0x40;

	static const ioaddr iCTR1_LOW = 0x41;
	static const ioaddr iCTR1_HIG = 0x41;
	static const ioaddr iSTS1_LOW = 0x41;
	static const ioaddr iSTS1_HIG = 0x41;

	static const ioaddr iCTR2_LOW = 0x42;
	static const ioaddr iCTR2_HIG = 0x42;
	static const ioaddr iSTS2_LOW = 0x42;
	static const ioaddr iSTS2_HIG = 0x42;

	static const ioaddr iSPR = 0x61;

	void start(natw N);
}

namespace piix3 {

	const natb APICBASE = 0x80;
	const natb XBCS = 0x4e;
	const natl XBCS_ENABLE = (1U << 8);
	const natw VENDOR_ID = 0x8086;
	const natw DEVICE_ID = 0x7000;

	bool ioapic_enable(natb bus, natb dev, natb fun);
	bool ioapic_disable(natb bus, natb dev, natb fun);
}

namespace serial {

	const ioaddr iCOM1_RBR = 0x03F8;
	const ioaddr iCOM1_THR = 0x03F8;
	const ioaddr iCOM1_LSR = 0x03FD;
	const ioaddr iCOM1_LCR = 0x03FB;
	const ioaddr iCOM1_DLR_LSB = 0x03F8;
	const ioaddr iCOM1_DLR_MSB = 0x03F9;
	const ioaddr iCOM1_IER = 0x03F9;
	const ioaddr iCOM1_MCR = 0x03FC;
	const ioaddr iCOM1_IIR = 0x03FA;

	const ioaddr iCOM2_RBR = 0x02F8;
	const ioaddr iCOM2_THR = 0x02F8;
	const ioaddr iCOM2_LSR = 0x02FD;
	const ioaddr iCOM2_LCR = 0x02FB;
	const ioaddr iCOM2_DLR_LSB = 0x02F8;
	const ioaddr iCOM2_DLR_MSB = 0x02F9;
	const ioaddr iCOM2_IER = 0x02F9;
	const ioaddr iCOM2_MCR = 0x02FC;
	const ioaddr iCOM2_IIR = 0x02FA;

}

namespace rnd {
	extern natl seed;
};
/// @endcond
